import time
import io
from zipfile import ZipFile
from dataclasses import dataclass, asdict
from contextlib import contextmanager
import json
from pathlib import Path

import requests
import pandas as pd

from toargridding.metadata import Metadata, AnalysisRequestResult, Coordinates


STATION_LAT = "station_coordinates_lat"
STATION_LON = "station_coordinates_lng"
COORDS = [STATION_LAT, STATION_LON]


@dataclass(frozen=True)
class QueryOptions:
    daterange: str
    variable_id: str
    statistics: str
    sampling: str
    min_data_capture: str = "0"  # TODO check effect on NaNs
    metadata_scheme: str = "basic"
    limit: str = "None"
    format: str = "by_statistic"

    @staticmethod
    def from_metadata(metadata: Metadata):
        return QueryOptions(
            daterange=metadata.time.daterange_option,
            variable_id=str(metadata.variable.toar_id),
            statistics=metadata.statistic,
            sampling=metadata.time.sampling,
        )

    @property
    def cache_key(self):
        return "".join(asdict(self).values())


class Cache:
    def __init__(self, cache_dir):
        self.cache_file = cache_dir / "status_endpoints.json"

        if not self.cache_file.is_file():  # initialize cache with dummy values
            with open(self.cache_file, "w") as cache:
                json.dump({"foo": "bar"}, cache)

    def __contains__(self, item: str):
        with self.storage_dict() as storage:
            return item in storage.keys()

    def get(self, key: str):
        with self.storage_dict() as storage:
            return storage[key]

    def put(self, key: str, content: str):
        with self.storage_dict() as storage:
            storage[key] = content

    def remove(self, key: str):
        with self.storage_dict() as storage:
            del storage[key]

    @contextmanager
    def storage_dict(self):
        with open(self.cache_file, "r") as cache:  # setup
            storage_dict = json.load(cache)

        yield storage_dict

        with open(self.cache_file, "w") as cache:  # teardown
            json.dump(storage_dict, cache)


class Connection:
    def __init__(self, endpoint, cache_dir):
        self.endpoint = endpoint
        self.cache = Cache(cache_dir)
        # max wait time is 30min
        self.wait_seconds = [minutes * 60 for minutes in (1, 2, 2, 5, 10, 10)]

    def get(self, query_options):
        status_endpoint = self.get_status_endpoint(query_options)

        for i, wait_time in enumerate(self.wait_seconds):
            print(f"try: {i+1}, wait_time: {wait_time}")
            response = self.wait_and_get(status_endpoint, wait_secs=wait_time)
            if response.headers == "application/zip":
                return response
        else:
            raise RuntimeError(
                f"No data available after {sum(self.wait_seconds) / 60} minutes. retry later."
            )

    def get_status_endpoint(self, query_options: QueryOptions):
        if query_options.cache_key in self.cache:
            status_endpoint = self.cache.get(query_options.cache_key)

            try:  # test for stale cache
                self.wait_and_get(status_endpoint).raise_for_status()
            except requests.exceptions.HTTPError:
                self.cache.remove(query_options.cache_key)
            else:
                print("load status endpoint from cache")
                return status_endpoint
        else:
            print("query not in cache")

        status_endpoint = self.query_for_status_endpoint(query_options)
        return status_endpoint

    def query_for_status_endpoint(self, query_options: QueryOptions):
        response = self.wait_and_get(self.endpoint, asdict(query_options))
        status_endpoint = response.json()["status"]
        self.cache.put(query_options.cache_key, status_endpoint)

        return status_endpoint

    def wait_and_get(
        self, endpoint, query_options=None, wait_secs=None, timeout=(3.05, 5)
    ):
        if wait_secs:
            time.sleep(wait_secs)

        return requests.get(endpoint, params=query_options, timeout=timeout)


class AnalysisService:
    METADATA = "metadata"

    def __init__(self, stats_endpoint, cache_dir):
        self.connection = Connection(stats_endpoint, cache_dir)

    def get_data(self, metadata: Metadata) -> AnalysisRequestResult:
        timeseries, timeseries_metadata = self.get_timeseries_and_metadata(metadata)
        coords = self.get_clean_coords(timeseries_metadata)
        timeseries = self.get_clean_timeseries(timeseries, metadata)
        return AnalysisRequestResult(timeseries, coords, metadata)

    def get_timeseries_and_metadata(self, metadata: Metadata):
        query_options = QueryOptions.from_metadata(metadata)
        content = self.connection.get(query_options)
        timeseries, timeseries_metadata = self.load_data(content, metadata)
        return timeseries, timeseries_metadata

    def get_clean_coords(self, timeseries_metadata):
        coords = timeseries_metadata[COORDS]
        coords.columns = [Coordinates.latitude.name, Coordinates.longitude.name]
        valid_coords = coords.notna().all(axis=1)
        return coords[valid_coords]

    def get_clean_timeseries(self, timeseries, metadata: Metadata):
        # TODO maybe use cf-index here already ?
        first, last = timeseries.columns[0], timeseries.columns[-1]
        # remove data where utc -> sun/local ? time conversion leads to dateshift
        timeseries.drop(columns=[first, last], inplace=True)
        timeseries.columns = metadata.time.as_datetime_index()

        all_na = timeseries.isna().all(axis=1)
        timeseries = timeseries[~all_na]
        timeseries = timeseries.fillna(0)

        return timeseries

    def load_data(
        self, content: bytes, metadata: Metadata
    ) -> (pd.DataFrame, pd.DataFrame):
        zip_stream = io.BytesIO(content)
        with ZipFile(zip_stream) as myzip:
            timeseries = self.extract_data(myzip, metadata.statistic)
            timeseries_metadata = self.extract_data(myzip, AnalysisService.METADATA)

        return timeseries, timeseries_metadata

    def extract_data(self, zip_file, data_file) -> pd.DataFrame:
        with zip_file.open(f"{data_file}.csv") as f:
            s_stream = io.StringIO(f.read().decode("utf-8"))
            return pd.read_csv(s_stream, comment="#", index_col=0)


class AnalysisServiceDownload(AnalysisService):
    def __init__(
        self, stats_endpoint, cache_dir, sample_dir: Path, use_downloaded=True
    ):
        super().__init__(stats_endpoint, cache_dir)
        self.sample_dir = sample_dir
        self.use_downloaded = use_downloaded

    def get_timeseries_and_metadata(self, metadata: Metadata):
        filename = self.sample_dir / self.get_sample_file_name(metadata)
        query_options = QueryOptions.from_metadata(metadata)
        needs_fresh_download = (not self.use_downloaded) or (not filename.is_file())

        if needs_fresh_download:
            response = self.connection.get(query_options)
            with open(filename, "w+b") as downloaded_file:
                downloaded_file.write(response.content)

        with open(filename, "r+b") as data_file:
            content = data_file.read()

        timeseries, timeseries_metadata = self.load_data(content, metadata)
        return timeseries, timeseries_metadata

    @staticmethod
    def get_sample_file_name(metadata: Metadata):
        return f"{metadata.statistic}_{metadata.time.sampling}_{metadata.time.start.date()}_{metadata.time.end.date()}.zip"
